@@ -29,9 +29,10 @@ import unittest
 import warnings
 from distutils.dir_util import copy_tree, remove_tree
 from pathlib import Path
-from unittest import TestCase
+from unittest import TestCase, mock
 
 import pyfakefs.tests.import_as_example
+import pyfakefs.tests.logsio
 from pyfakefs import fake_filesystem_unittest, fake_filesystem
 from pyfakefs.extra_packages import pathlib
 from pyfakefs.fake_filesystem import OSType
@@ -50,13 +51,35 @@ class TestPatcher(TestCase):
             self.assertEqual('test', contents)
 
     @patchfs
-    def test_context_decorator(self, fs):
-        fs.create_file('/foo/bar', contents='test')
+    def test_context_decorator(self, fake_fs):
+        fake_fs.create_file('/foo/bar', contents='test')
         with open('/foo/bar') as f:
             contents = f.read()
         self.assertEqual('test', contents)
 
 
+class TestPatchfsArgumentOrder(TestCase):
+    @patchfs
+    @mock.patch('os.system')
+    def test_argument_order1(self, fake_fs, patched_system):
+        fake_fs.create_file('/foo/bar', contents='test')
+        with open('/foo/bar') as f:
+            contents = f.read()
+        self.assertEqual('test', contents)
+        os.system("foo")
+        patched_system.assert_called_with("foo")
+
+    @mock.patch('os.system')
+    @patchfs
+    def test_argument_order2(self, patched_system, fake_fs):
+        fake_fs.create_file('/foo/bar', contents='test')
+        with open('/foo/bar') as f:
+            contents = f.read()
+        self.assertEqual('test', contents)
+        os.system("foo")
+        patched_system.assert_called_with("foo")
+
+
 class TestPyfakefsUnittestBase(fake_filesystem_unittest.TestCase):
     def setUp(self):
         """Set up the fake file system"""
@@ -201,7 +224,10 @@ class TestPatchingImports(TestPyfakefsUnittestBase):
         self.assertEqual('abc', contents)
 
 
-class TestPatchingDefaultArgs(TestPyfakefsUnittestBase):
+class TestPatchingDefaultArgs(fake_filesystem_unittest.TestCase):
+    def setUp(self):
+        self.setUpPyfakefs(patch_default_args=True)
+
     def test_path_exists_as_default_arg_in_function(self):
         file_path = '/foo/bar'
         self.fs.create_dir(file_path)
@@ -214,6 +240,11 @@ class TestPatchingDefaultArgs(TestPyfakefsUnittestBase):
         sut = pyfakefs.tests.import_as_example.TestDefaultArg()
         self.assertTrue(sut.check_if_exists(file_path))
 
+    def test_fake_path_exists4(self):
+        self.fs.create_file('foo')
+        self.assertTrue(
+            pyfakefs.tests.import_as_example.check_if_exists4('foo'))
+
 
 class TestAttributesWithFakeModuleNames(TestPyfakefsUnittestBase):
     """Test that module attributes with names like `path` or `io` are not
@@ -294,11 +325,6 @@ class NoSkipNamesTest(fake_filesystem_unittest.TestCase):
         self.assertTrue(
             pyfakefs.tests.import_as_example.check_if_exists3('foo'))
 
-    def test_fake_path_exists4(self):
-        self.fs.create_file('foo')
-        self.assertTrue(
-            pyfakefs.tests.import_as_example.check_if_exists4('foo'))
-
     def test_fake_path_exists5(self):
         self.fs.create_file('foo')
         self.assertTrue(
@@ -318,6 +344,13 @@ class NoSkipNamesTest(fake_filesystem_unittest.TestCase):
         with self.assertRaises(OSError):
             pyfakefs.tests.import_as_example.open_this_file()
 
+    def test_open_patched_in_module_ending_with_io(self):
+        # regression test for #569
+        file_path = '/foo/bar'
+        self.fs.create_file(file_path, contents=b'abc')
+        contents = pyfakefs.tests.logsio.file_contents(file_path)
+        self.assertEqual(b'abc', contents)
+
 
 class AdditionalSkipNamesTest(fake_filesystem_unittest.TestCase):
     """Make sure that modules in additional_skip_names are not patched.
@@ -468,17 +501,17 @@ class PatchModuleTestUsingDecorator(unittest.TestCase):
 
     @patchfs
     @unittest.expectedFailure
-    def test_system_stat_failing(self, fs):
+    def test_system_stat_failing(self, fake_fs):
         file_path = '/foo/bar'
-        fs.create_file(file_path, contents=b'test')
+        fake_fs.create_file(file_path, contents=b'test')
         self.assertEqual(
             4, pyfakefs.tests.import_as_example.system_stat(file_path).st_size)
 
     @patchfs(modules_to_patch={
         'pyfakefs.tests.import_as_example': FakeExampleModule})
-    def test_system_stat(self, fs):
+    def test_system_stat(self, fake_fs):
         file_path = '/foo/bar'
-        fs.create_file(file_path, contents=b'test')
+        fake_fs.create_file(file_path, contents=b'test')
         self.assertEqual(
             4, pyfakefs.tests.import_as_example.system_stat(file_path).st_size)
 
@@ -698,6 +731,8 @@ class PathlibTest(TestCase):
 
 
 class TestDeprecationSuppression(fake_filesystem_unittest.TestCase):
+    @unittest.skipIf(sys.version_info[1] == 6,
+                     'Test fails for Python 3.6 for unknown reason')
     def test_no_deprecation_warning(self):
         """Ensures that deprecation warnings are suppressed during module
         lookup, see #542.
